# SPDX-FileCopyrightText: 2025 rakurtz, James R. Barlow
# SPDX-License-Identifier: CC0-1.0
from __future__ import annotations

import pytest

from pikepdf import Matrix, Pdf, get_objects_with_ctm
from pikepdf.models.ctm import MatrixStack


# Crafted test PDF that contains 4 'Do' (Draw Object) operations
# However only three of those have a valid CTM
# First: 'Do' without previous 'cm' operation
# Second: 'Do' with 'cm' operation (100 0 0 100 100 100 cm)
# Third: 'Do' preceed by invalid 'cm' operation (0 cm)
# Fourth: 'Do' after invalid CTM has been popped off stack
@pytest.fixture
def ctm_cm(resources):
    with Pdf.open(resources / 'ctm_cm.pdf') as pdf:
        yield pdf


def test_get_matrices(ctm_cm):
    matrixes = [matrix for _, matrix in get_objects_with_ctm(ctm_cm.pages[0])]

    first = Matrix(1, 0, 0, 1, 0, 0)  # identity matrix
    second = Matrix(100, 0, 0, 100, 100, 100)
    # third drawn object doesn't have a valid CTM and is skipped
    fourth = Matrix(2, 0, 0, 2, 2, 2)

    assert len(matrixes) == 3
    assert matrixes[0] == first
    assert matrixes[1] == second
    assert matrixes[2] == second @ fourth


def test_get_matrices_scaled(ctm_cm):
    scale = Matrix(7, 0, 0, 7, 0, 0)
    matrixes = [matrix for _, matrix in get_objects_with_ctm(ctm_cm.pages[0], scale)]

    first = scale
    second = Matrix(700, 0, 0, 700, 100, 100)
    # third drawn object doesn't have a valid CTM and is skipped
    fourth = Matrix(2, 0, 0, 2, 2, 2)

    assert len(matrixes) == 3
    assert matrixes[0] == first
    assert matrixes[1] == second
    assert matrixes[2] == second @ fourth


def test_underflow():
    stack = MatrixStack()
    for _ in range(4):
        stack.pop()
    assert stack.ctm == Matrix.identity()


def test_invalid_ctm():
    stack = MatrixStack()
    stack.invalidate_current_transformation_matrix()
    assert stack.ctm is None



# SPDX-FileCopyrightText: 2022 James R. Barlow
# SPDX-License-Identifier: CC0-1.0

from __future__ import annotations

import pytest

from pikepdf import Annotation, Name, Pdf


@pytest.fixture
def form(resources):
    with Pdf.open(resources / 'form.pdf') as pdf:
        yield pdf


@pytest.fixture
def dd0293(resources):
    """This is a real-world fillable form with significantly more complexity than the
    basic example form above. It has examples of:

    * Choice fields
    * Digital signature fields
    * Multi-line text fields
    * Fields with an alternate name
    """
    with Pdf.open(resources / 'form_dd0293.pdf') as pdf:
        yield pdf


def test_acroform(form):
    acro = form.acroform
    assert acro.exists
    assert len(acro.fields) == 5


def test_appearances(form):
    acro = form.acroform
    acro.needs_appearances = True
    assert acro.needs_appearances is True
    acro.generate_appearances_if_needed()
    assert acro.needs_appearances is False


def test_text(form):
    field = form.acroform.get_fields_with_qualified_name('Text1')[0]
    assert field.fully_qualified_name == 'Text1'
    assert field.default_appearance == b"/Helv 12 Tf 0 g"
    assert field.get_inheritable_field_value('/DA') == b"/Helv 12 Tf 0 g"
    assert field.get_inheritable_field_value_as_string('/DA') == "/Helv 12 Tf 0 g"
    assert Name.Font in field.default_resources
    assert field.top_level_field == field
    assert field.is_text
    assert not field.is_checkbox
    assert not field.is_radio_button
    assert not field.is_pushbutton
    assert not field.is_choice


def test_button(form):
    field = form.acroform.get_fields_with_qualified_name('Button2')[0]
    assert field.fully_qualified_name == 'Button2'
    assert field.top_level_field == field
    assert not field.is_text
    assert not field.is_checkbox
    assert not field.is_radio_button
    assert field.is_pushbutton
    assert not field.is_choice


def test_checkbox(form):
    field = form.acroform.get_fields_with_qualified_name('Check Box3')[0]
    assert field.fully_qualified_name == 'Check Box3'
    assert field.top_level_field == field
    assert not field.is_text
    assert field.is_checkbox
    assert not field.is_radio_button
    assert not field.is_pushbutton
    assert not field.is_choice


def test_radio_button(form):
    fields = form.acroform.get_fields_with_qualified_name('Group4')
    assert len(fields) == 1  # 1 group, but not the 2 individual radio buttons
    top_field = fields[0]
    assert top_field.fully_qualified_name == 'Group4'
    assert top_field.top_level_field == top_field
    assert not top_field.is_text
    assert not top_field.is_checkbox
    assert top_field.is_radio_button
    assert not top_field.is_pushbutton
    assert not top_field.is_choice
    terminal_fields = list(
        filter(lambda f: f.fully_qualified_name == 'Group4', form.acroform.fields)
    )
    assert len(terminal_fields) == 2
    for field in terminal_fields:
        assert field.top_level_field != field
        assert field.parent == top_field
        assert field.top_level_field == top_field
        assert field.get_inheritable_field_value_as_name('/FT') == top_field.field_type
        assert not field.is_text
        assert not field.is_checkbox
        assert field.is_radio_button
        assert not field.is_pushbutton
        assert not field.is_choice


def test_choice(dd0293):
    field = dd0293.acroform.get_fields_with_qualified_name(
        'form1[0].page1[0].#subform[2].DropDownList1[0]'
    )[0]
    assert (
        field.fully_qualified_name == 'form1[0].page1[0].#subform[2].DropDownList1[0]'
    )
    assert field.partial_name == 'DropDownList1[0]'
    assert (
        field.alternate_name
        == 'SECTION 2: SERVICE INFORMATION (Information from DD Form 214. Include Member Copy of DD Form 214 and enter as much as is readily available. - 7. GRADE/RANK AT DISCHARGE - ARMY - Select from drop-down list.'
    )
    assert field.top_level_field.partial_name == 'form1[0]'
    assert not field.is_text
    assert not field.is_checkbox
    assert not field.is_radio_button
    assert not field.is_pushbutton
    assert field.is_choice
    # In theory, field.choices should be populated. However, this does not
    # work. The underlaying QPDF implementation appears to be at fault. See
    # https://github.com/qpdf/qpdf/issues/1433


def test_existing_value(dd0293):
    field = dd0293.acroform.get_fields_with_qualified_name(
        'form1[0].#pageSet[0].Page1[0].TextField7[0]'
    )[0]
    assert field.value == "Controlled by: CUI Category: LDC: POC: "
    assert field.value_as_string == "Controlled by: CUI Category: LDC: POC: "
    assert field.default_value == "Controlled by: \nCUI Category: \nLDC: \nPOC: "
    assert (
        field.default_value_as_string == "Controlled by: \nCUI Category: \nLDC: \nPOC: "
    )


def test_remove_fields(form):
    acro = form.acroform
    fields = acro.get_fields_with_qualified_name('Button2')
    assert len(fields) == 1
    acro.remove_fields(fields)
    fields = acro.get_fields_with_qualified_name('Button2')
    assert len(fields) == 0


def test_disable_signatures(dd0293):
    sigs = [f for f in dd0293.pages[1].Annots if hasattr(f, 'FT') and f.FT == '/Sig']
    assert len(sigs) == 1
    dd0293.acroform.disable_digital_signatures()
    sigs = [f for f in dd0293.pages[1].Annots if hasattr(f, 'FT') and f.FT == '/Sig']
    assert len(sigs) == 0


def test_get_annotations_for_field(form):
    field = form.acroform.get_fields_with_qualified_name('Text1')[0]
    annots = form.acroform.get_annotations_for_field(field)
    assert len(annots) == 1
    print(annots[0])
    assert annots[0].subtype == Name.Widget


def test_get_widget_annotations_for_page(form):
    page = form.pages[0]
    annots = form.acroform.get_widget_annotations_for_page(page)
    assert (
        len(annots) == 5
    )  # 1 annotation per terminal field (radio buttons each have their own)


def test_get_form_fields_for_page(form):
    page = form.pages[0]
    fields = form.acroform.get_form_fields_for_page(page)
    assert len(fields) == 4  # 1 field per top-level field (2 radio buttons count as 1)


def test_get_field_for_annotation(form):
    annot = Annotation(form.Root.AcroForm.Fields[1])
    field = form.acroform.get_field_for_annotation(annot)
    assert field.fully_qualified_name == 'Button2'


def test_copy_form(form, dd0293):
    orig_count = len(dd0293.acroform.fields)
    dd0293.pages.extend(form.pages)
    copied_fields = dd0293.acroform.fix_copied_annotations(
        dd0293.pages[-1], form.pages[0], form.acroform
    )
    new_count = len(dd0293.acroform.fields)
    assert len(copied_fields) == 4  # Count is top-level fields
    assert new_count == orig_count + 5  # Count is terminal fields



# SPDX-FileCopyrightText: 2025 Dominick Johnson
# SPDX-License-Identifier: CC0-1.0

from __future__ import annotations

import pytest

from pikepdf import Name, Pdf
from pikepdf.form import (
    CheckboxField,
    ChoiceField,
    DefaultAppearanceStreamGenerator,
    ExtendedAppearanceStreamGenerator,
    Form,
    RadioButtonGroup,
    TextField,
)


@pytest.fixture
def form(resources):
    with Pdf.open(resources / 'form.pdf') as pdf:
        yield pdf


@pytest.fixture
def dd0293(resources):
    """This is a real-world fillable form with significantly more complexity than the
    basic example form above. It has examples of:

    * Choice fields
    * Digital signature fields
    * Multi-line text fields
    * Fields with an alternate name
    """
    with Pdf.open(resources / 'form_dd0293.pdf') as pdf:
        yield pdf


@pytest.fixture
def va210966(resources):
    """This is a real-world fillable form with significantly more complexity than the
    basic example form above. It has examples of:

    * Combed text fields
    * Radio buttons with parent fields
    * Digital signature fields
    * Fields with an alternate name
    """
    with Pdf.open(resources / 'form_210966.pdf') as pdf:
        yield pdf


def test_form_getitem(form):
    f = Form(form)
    assert isinstance(f['Text1'], TextField)
    assert isinstance(f['Check Box3'], CheckboxField)
    assert isinstance(f['Group4'], RadioButtonGroup)


def test_form_items(form):
    f = Form(form)
    d = dict(f.items())
    assert isinstance(d['Text1'], TextField)
    assert isinstance(d['Check Box3'], CheckboxField)
    assert isinstance(d['Group4'], RadioButtonGroup)


def test_text(form):
    f = Form(form)
    field = f['Text1']
    assert field.is_multiline is False
    assert field.is_required is False
    assert field.value == ''
    field.value = 'Stuff'
    assert field.value == 'Stuff'
    assert f._acroform.needs_appearances


def test_checkbox(form):
    f = Form(form)
    field = f['Check Box3']
    assert field.checked is False
    assert field.on_value == Name.Yes
    field.checked = True
    assert field.checked is True
    assert field._field.obj.AS == Name.Yes
    field.checked = False
    assert field.checked is False


def test_radio(form):
    f = Form(form)
    field = f['Group4']
    assert field.value is None
    assert field.selected is None
    assert len(field.states) == 2
    assert len(field.options) == 2
    assert Name('/Choice1') in field.states
    assert Name('/Choice2') in field.states
    assert field.options[0].on_value == Name('/Choice1')
    assert field.options[1].on_value == Name('/Choice2')
    assert not field.options[0].checked
    assert not field.options[1].checked
    # Set value directly
    field.value = Name('/Choice1')
    assert field.value == Name('/Choice1')
    assert field.options[0].checked
    assert not field.options[1].checked
    assert field.options[0]._annot_dict.AS == Name('/Choice1')
    assert field.options[1]._annot_dict.AS == Name('/Off')
    # Using `group.selected`
    field.selected = field.options[1]
    assert field.value == Name('/Choice2')
    assert not field.options[0].checked
    assert field.options[1].checked
    assert field.options[0]._annot_dict.AS == Name('/Off')
    assert field.options[1]._annot_dict.AS == Name('/Choice2')
    # Using `option.select`
    field.options[0].select()
    assert field.value == Name('/Choice1')
    assert field.options[0].checked
    assert not field.options[1].checked
    assert field.options[0]._annot_dict.AS == Name('/Choice1')
    assert field.options[1]._annot_dict.AS == Name('/Off')
    # Using `option.checked`
    field.options[1].checked = True
    assert field.value == Name('/Choice2')
    assert not field.options[0].checked
    assert field.options[1].checked
    assert field.options[0]._annot_dict.AS == Name('/Off')
    assert field.options[1]._annot_dict.AS == Name('/Choice2')
    # We don't support directly unchecking a radio button
    with pytest.raises(ValueError):
        field.options[1].checked = False


def test_radio_with_parent(va210966):
    f = Form(va210966)
    field = f['F[0].Page_1[0].RadioButtonList[1]']
    assert field.value == Name.Off
    assert field.selected is None
    assert len(field.states) == 7
    assert len(field.options) == 7
    assert Name('/0') in field.states
    assert Name('/6') in field.states
    assert field.options[0].on_value == Name('/0')
    assert field.options[6].on_value == Name('/6')
    assert not field.options[0].checked
    assert not field.options[6].checked
    # Set value directly
    field.value = Name('/0')
    assert field.value == Name('/0')
    assert field.options[0].checked
    assert field.options[0]._annot_dict.AS == Name('/0')
    # Using `group.selected`
    field.selected = field.options[1]
    assert field.value == Name('/1')
    assert not field.options[0].checked
    assert field.options[1].checked
    assert field.options[0]._annot_dict.AS == Name('/Off')
    assert field.options[1]._annot_dict.AS == Name('/1')
    # Using `option.select`
    field.options[2].select()
    assert field.value == Name('/2')
    assert field.options[2].checked
    assert not field.options[1].checked
    assert field.options[2]._annot_dict.AS == Name('/2')
    assert field.options[1]._annot_dict.AS == Name('/Off')
    # Using `option.checked`
    field.options[3].checked = True
    assert field.value == Name('/3')
    assert not field.options[2].checked
    assert field.options[3].checked
    assert field.options[2]._annot_dict.AS == Name('/Off')
    assert field.options[3]._annot_dict.AS == Name('/3')


def test_choice(dd0293):
    f = Form(dd0293)
    field = f['form1[0].page1[0].#subform[2].DropDownList1[5]']
    assert isinstance(field, ChoiceField)
    assert field.is_combobox
    assert not field.is_multiselect
    assert not field.allow_edit
    assert field.value is None
    assert field.options[1].display_value == 'SPC1/E-1'
    assert field.options[1].export_value == 'SPC1/E-1'
    assert not field.options[1].selected
    field.value = 'SPC1/E-1'
    assert field.value == 'SPC1/E-1'
    assert f._acroform.needs_appearances
    assert field.options[1].selected
    with pytest.raises(ValueError):
        field.value = 'PVT/E-1'
    field.options[2].select()
    assert field.value == 'SPC2/E-2'


def test_signature_stamp(resources, dd0293):
    f = Form(dd0293)
    field = f['form1[0].page2[0].SignatureField1[0]']
    with Pdf.open(resources / 'pike-jp2.pdf') as sig_pdf:
        xobj_name = field.stamp_overlay(sig_pdf.pages[0])
    assert xobj_name in dd0293.pages[1].Resources.XObject
    stream = dd0293.pages[1].Contents.read_bytes()
    assert bytes(xobj_name) + b' Do' in stream


def test_signature_stamp_expand(resources, va210966):
    f = Form(va210966)
    field = f['F[0].#subform[1].Digital_Signature[0]']
    with Pdf.open(resources / 'pike-jp2.pdf') as sig_pdf:
        xobj_name = field.stamp_overlay(sig_pdf.pages[0], expand_rect=(0, 17))
    assert xobj_name in va210966.pages[1].Resources.XObject
    stream = va210966.pages[1].Contents.read_bytes()
    assert bytes(xobj_name) + b' Do' in stream


def test_default_appearance_generator_text(form):
    f = Form(form, DefaultAppearanceStreamGenerator)
    field = f['Text1']
    field.value = 'Stuff'
    assert field.value == 'Stuff'
    assert not f._acroform.needs_appearances
    stream = field._field.obj.AP.N.read_bytes()
    assert field._field.default_appearance in stream
    assert b"(Stuff)" in stream


def test_extended_appearance_generator_multiline_text(dd0293):
    text = (
        "Manual Break:\n"
        "This is a really long line designed to trigger a line break in the word-wrap "
        "code. It should totally be long enough to break to the second line. If not, "
        "we'll just type some more nonsense. Nonsense is actually sometimes quite "
        "useful. You might think I'm stalling, but actually, I'm filibustering. They're "
        "totally different things. *Wink* *Nod* *Nod off*"
    )
    f = Form(dd0293, ExtendedAppearanceStreamGenerator)
    field = f['form1[0].page2[0].TextField4[0]']
    field.value = text
    assert field.value == text
    assert not f._acroform.needs_appearances
    stream = field._field.obj.AP.N.read_bytes()
    assert field._field.default_appearance in stream
    assert b"Manual" in stream
    assert b"nonsense" in stream


def test_extended_appearance_generator_combed_text(va210966):
    f = Form(va210966, ExtendedAppearanceStreamGenerator)
    field = f['F[0].Page_1[0].Veterans_First_Name[0]']
    field.value = 'Nemo'
    assert field.value == 'Nemo'
    assert not f._acroform.needs_appearances
    stream = field._field.obj.AP.N.read_bytes()
    assert field._field.default_appearance in stream
    assert b"(N)" in stream
    assert b"(e)" in stream
    assert b"(m)" in stream
    assert b"(o)" in stream
    assert b"] TJ" in stream



# SPDX-FileCopyrightText: 2022 James R. Barlow
# SPDX-License-Identifier: CC0-1.0

from __future__ import annotations

from decimal import Decimal

import pytest
from PIL import Image

from pikepdf import Matrix, Pdf
from pikepdf.canvas import (
    BLACK,
    Canvas,
    Color,
    ContentStreamBuilder,
    Helvetica,
    SimpleFont,
    Text,
    TextDirection,
)
from pikepdf.objects import Name, Operator


@pytest.fixture
def pdf(request, resources):
    with Pdf.open(resources / request.param) as pdf:
        yield pdf


@pytest.fixture
def resource_dict(request, pdf):
    if request.param == 'form':
        return pdf.Root.AcroForm.DR
    else:
        return pdf.pages[0].Resources


@pytest.fixture
def simplefont(request, resource_dict):
    return SimpleFont.load(request.param, resource_dict)


FONTS = {
    # TODO These are all TrueType fonts using the WinAnsiEncoding. Would be good to find
    # other examples. Having trouble finding other example PDFs with fonts that would be
    # supported by our current limited implementation. All the Type1 fonts I see reference
    # built-in fonts with no widths, or have difference maps, and MacRoman doesn't seem to
    # really exist in the wild anymore as far as I can tell.
    'arial-truetype-winansi': ('form_dd0293.pdf', 'form', Name('/ArialMT')),
    'arialbold-truetype-winansi': ('form_dd0293.pdf', 'form', Name('/Arial-BoldMT')),
    'times-truetype-winansi': ('form_dd0293.pdf', 'form', Name('/TimesNewRomanPSMT')),
}


class TestSimpleFont:
    @pytest.mark.parametrize(
        'pdf,resource_dict,name',
        [
            FONTS['arial-truetype-winansi'],
            FONTS['arialbold-truetype-winansi'],
            FONTS['times-truetype-winansi'],
        ],
        indirect=['pdf', 'resource_dict'],
    )
    def test_load(self, pdf, resource_dict, name):
        font = SimpleFont.load(name, resource_dict)
        assert font.data == resource_dict.Font[name]

    @pytest.mark.parametrize(
        'pdf,resource_dict,simplefont,char_code,expected_width',
        [
            (*FONTS['arial-truetype-winansi'], ord('a'), 556),
            (*FONTS['arialbold-truetype-winansi'], b'a', 556),
            (*FONTS['times-truetype-winansi'], 'a', 444),
        ],
        indirect=['pdf', 'resource_dict', 'simplefont'],
    )
    def test_unscaled_char_width_known(
        self, pdf, resource_dict, simplefont, char_code, expected_width
    ):
        width = simplefont.unscaled_char_width(char_code)
        assert width == expected_width

    @pytest.mark.parametrize(
        'pdf,resource_dict,simplefont',
        [
            FONTS['arial-truetype-winansi'],
            FONTS['arialbold-truetype-winansi'],
            FONTS['times-truetype-winansi'],
        ],
        indirect=True,
    )
    @pytest.mark.parametrize(
        'font_size,width,expected',
        [
            (1, 1000, 1),
            (12, 500, 6),
        ],
    )
    def test_convert_width(
        self, pdf, resource_dict, simplefont, font_size, width, expected
    ):
        assert simplefont.convert_width(width, font_size) == expected

    @pytest.mark.parametrize(
        'pdf,resource_dict,simplefont,string,encoded',
        [
            (
                *FONTS['arial-truetype-winansi'],
                "This is just ASCII!",
                b"This is just ASCII!",
            ),
            (
                *FONTS['times-truetype-winansi'],
                # Sentence constructed in a deliberate attempt to use multiple different
                # non-ascii latin characters, don't read too much into it. I assure you my
                # grandmother and I are fine.
                "«Disparaître avec ma grand-mère française aiguë à l'hôpital dégoûtant cet été»",
                b"\253Dispara\356tre avec ma grand-m\350re fran\347aise aigu\353 \340 l'h\364pital d\351go\373tant cet \351t\351\273",
            ),
        ],
        indirect=['pdf', 'resource_dict', 'simplefont'],
    )
    def test_unscaled_char_width_known2(
        self, pdf, resource_dict, simplefont, string, encoded
    ):
        assert simplefont.encode(string) == encoded


class TestContentStreamBuilder:
    def test_init(self):
        builder = ContentStreamBuilder()
        assert isinstance(builder.build(), bytes)

    def test_append(self):
        builder = ContentStreamBuilder()
        builder.push()
        assert builder.build() == b'q\n'

    def test_extend(self):
        builder1 = ContentStreamBuilder()
        builder2 = ContentStreamBuilder()
        builder2.push()
        builder1.extend(builder2)
        assert builder1.build() == b'q\n'

    @pytest.mark.parametrize(
        'method,args,operator',
        [
            (ContentStreamBuilder.push, (), 'q'),
            (ContentStreamBuilder.pop, (), 'Q'),
            (ContentStreamBuilder.cm, (Matrix(),), 'cm'),
            (
                ContentStreamBuilder.begin_marked_content_proplist,
                (Name.Test, 42),
                'BDC',
            ),
            (ContentStreamBuilder.end_marked_content, (), 'EMC'),
            (ContentStreamBuilder.begin_marked_content, (Name.Foo,), 'BMC'),
            (ContentStreamBuilder.begin_text, (), 'BT'),
            (ContentStreamBuilder.end_text, (), 'ET'),
            (ContentStreamBuilder.set_text_font, (Name.Test, 12), 'Tf'),
            (ContentStreamBuilder.set_text_font, (Name.Test, 12.5), 'Tf'),
            (ContentStreamBuilder.set_text_font, (Name.Test, Decimal('12.5')), 'Tf'),
            (ContentStreamBuilder.set_text_char_spacing, (1,), 'Tc'),
            (ContentStreamBuilder.set_text_char_spacing, (0.5,), 'Tc'),
            (ContentStreamBuilder.set_text_char_spacing, (Decimal('1'),), 'Tc'),
            (ContentStreamBuilder.set_text_word_spacing, (1,), 'Tw'),
            (ContentStreamBuilder.set_text_word_spacing, (0.5,), 'Tw'),
            (ContentStreamBuilder.set_text_word_spacing, (Decimal('1'),), 'Tw'),
            (ContentStreamBuilder.set_text_leading, (13,), 'TL'),
            (ContentStreamBuilder.set_text_leading, (13.5,), 'TL'),
            (ContentStreamBuilder.set_text_leading, (Decimal('13.5'),), 'TL'),
            (ContentStreamBuilder.set_text_matrix, (Matrix(),), "Tm"),
            (ContentStreamBuilder.set_text_rendering, (3,), "Tr"),
            (ContentStreamBuilder.set_text_horizontal_scaling, (100.0,), "Tz"),
            (
                ContentStreamBuilder.show_text_with_kerning,
                (b'A', 120, b'W', 120, b'A', 95, b'Y'),
                "TJ",
            ),
            (ContentStreamBuilder.show_text_line, (b'hello world',), "'"),
            (
                ContentStreamBuilder.show_text_line_with_spacing,
                (b'hello world', 0.25, 0.25),
                '"',
            ),
            (ContentStreamBuilder.move_cursor, (1, 2), "Td"),
            (ContentStreamBuilder.move_cursor_new_line, (), "T*"),
            (ContentStreamBuilder.stroke_and_close, (), "s"),
            (ContentStreamBuilder.fill, (), "f"),
            (ContentStreamBuilder.append_rectangle, (10, 10, 40, 40), "re"),
            (ContentStreamBuilder.set_stroke_color, (1, 0, 1), "RG"),
            (ContentStreamBuilder.set_fill_color, (0, 1, 0), "rg"),
            (ContentStreamBuilder.set_line_width, (5,), "w"),
            (ContentStreamBuilder.line, (1, 2, 3, 4), "l"),
            (ContentStreamBuilder.set_dashes, (), "d"),
            (ContentStreamBuilder.set_dashes, (1,), "d"),
            (ContentStreamBuilder.set_dashes, ([1, 2], 1), "d"),
            (ContentStreamBuilder.draw_xobject, (Name.X,), "Do"),
        ],
    )
    def test_operators(self, method, operator, args):
        builder = ContentStreamBuilder()
        method(builder, *args)
        assert builder.build().endswith(Operator(operator).unparse() + b'\n')


class TestCanvas:
    def test_basic(self):
        canvas = Canvas(page_size=(100, 100))
        assert canvas.page_size == (100, 100)
        with canvas.do.save_state(cm=Matrix().scaled(2, 2)):
            canvas.do.stroke_color(Color(1, 0, 0, 1)).line_width(2).dashes(1, 1).line(
                0, 0, 10, 10
            )
            canvas.do.fill_color(Color(0, 1, 0, 1)).rect(
                10, 10, 10, 10, fill=False
            ).rect(10, 10, 5, 5, fill=True)
        pdf = canvas.to_pdf()
        assert len(pdf.pages) == 1
        pdf.check_pdf_syntax()

    def test_image(self, resources):
        canvas = Canvas(page_size=(400, 100))
        canvas.do.draw_image(resources / 'pink-palette-icc.png', 0, 0, 100, 100)
        im = Image.open(resources / 'pink-palette-icc.png')
        canvas.do.draw_image(im.convert('1'), 100, 0, 100, 100)
        canvas.do.draw_image(im.convert('L'), 200, 0, 100, 100)
        canvas.do.draw_image(im.convert('RGB'), 300, 0, 100, 100)

        pdf = canvas.to_pdf()
        pdf.check_pdf_syntax()

    def test_text(self):
        hello_msg = 'Hello, World!'
        hello_arabic = 'مرحبا بالعالم'
        canvas = Canvas(page_size=(100, 100))

        text = Text()
        text.font(Name.Helvetica, 12).render_mode(1).text_transform(
            Matrix().translated(10, 10)
        ).horiz_scale(110).move_cursor(10, 10).show(hello_msg)

        # This is cheating! We're using one of the 14 base PDF fonts for a quick
        # test. If the resulting PDF is viewed, the result will not be in Arabic.
        # This does not properly register the font. The point of this test is
        # to ensure that the content stream is properly encoded.
        rtltext = Text(TextDirection.RTL)
        rtltext.font(Name.Helvetica, 12).render_mode(0).text_transform(
            Matrix().translated(10, 10)
        ).move_cursor(50, 50).show(hello_arabic)

        canvas.do.stroke_color(BLACK).draw_text(text)
        canvas.do.fill_color(BLACK).draw_text(rtltext)
        canvas.add_font(Name.Helvetica, Helvetica())
        pdf = canvas.to_pdf()
        pdf.check_pdf_syntax()

        for msg in [hello_msg, hello_arabic]:
            # str -> UTF-16 big endian bytes -> hex encoded str -> hex bytes
            hex_bytes = msg.encode('utf-16be').hex().encode('ascii')
            assert hex_bytes in pdf.pages[0].Contents.read_bytes()

    def test_stack_abuse(self, caplog):
        canvas = Canvas(page_size=(100, 100))
        canvas.do.pop().pop()
        canvas.to_pdf()
        assert "Graphics state stack is not empty when page saved" in caplog.text
